import babel from "enjoy-babel";
import asyncFunctions from "babel-plugin-syntax-async-functions";
import classProperties from "babel-plugin-syntax-class-properties";
import decorators from "babel-plugin-syntax-decorators";
import objectRestSpread from "babel-plugin-syntax-object-rest-spread";
import transformReactJsx from "babel-plugin-transform-react-jsx";
import pagePackage from "bundle-loader-enjoy-rn-html/src/page-package";
import transHtml from "./trans-html";

// 读取file中的内容，转换html标签，解析其中使用的组件列表
export default function (code, file, elementFactoryName, platform, callback){
	var components = {};

	var result = babel.transform(code.replace(/<\!\-\-[\s\S]*?\-\->/g, ""), {
		sourceFileName: file,
		plugins: [
			asyncFunctions,
			classProperties,
			decorators,
			objectRestSpread,
			[pagePackage, {
				platform: "web"
			}],
			[transformReactJsx, {
				"pragma": elementFactoryName
			}],
			[transHtml, {
				elementFactoryName: elementFactoryName,
				platform: platform
			}],
			function(){
				return {
					visitor: {
						CallExpression(path){
							var node = path.node;

							if(node.callee.type === "Identifier" &&
								node.callee.name === elementFactoryName &&
								node.arguments[0].type === "Identifier"){
								// 如果代码中没有定义该标签，则从enjoy导入
								if(!path.scope.hasBinding(node.arguments[0].name)){
									components[node.arguments[0].name] = true;
								}
							}
						}
					}
				};
			}
		]
	});

	components = Object.keys(components);

	if(components.length){
		components = components.join(", ") + ", ";
	}else{
		components = "";
	}

	callback(result, components);
}