import fs from "fs";
import path from "path";
import css2obj from "bundle-loader-enjoy-web-css/src/css-to-obj";

function toCamel(str){
	return str.replace(/\-[a-z]/g, function(char){
		return char.replace("-", "").toUpperCase();
	});
}

export default function({ types: t }) {
	return {
		visitor: {
			/**
			 * 将HTML标签转换为ReactNative标签，并添加默认样式，生成CSS选择器ID
			 */
			CallExpression: {
				enter(path, state){
					var node = path.node;
					var $arguments = node.arguments;

					if(node.callee.name === state.opts.elementFactoryName && $arguments[0].type === "StringLiteral"){
						let htmlTag;

						if($arguments[0].type === "StringLiteral"){
							htmlTag = $arguments[0].value;
						}else if($arguments[0].type === "Identifier"){
							htmlTag = $arguments[0].name;
						}else{
							return;
						}

						let reactTag;
						let isHtmlTag = false;

						switch(htmlTag){
							case "abbr":
							case "b":
							case "bdi":
							case "bdo":
							case "big":
							case "cite":
							case "code":
							case "del":
							case "dfn":
							case "em":
							case "i":
							case "ins":
							case "kbd":
							case "mark":
							case "meter":
							case "output":
							case "q":
							case "samp":
							case "small":
							case "span":
							case "strong":
							case "sub":
							case "sup":
							case "time":
							case "tt":
							case "var":
							case "address":
							case "article":
							case "aside":
							case "blockquote":
							case "dd":
							case "details":
							case "div":
							case "dl":
							case "dt":
							case "fieldset":
							case "figcaption":
							case "figure":
							case "footer":
							case "form":
							case "h1":
							case "h2":
							case "h3":
							case "h4":
							case "h5":
							case "h6":
							case "header":
							case "legend":
							case "li":
							case "main":
							case "menu":
							case "nav":
							case "ol":
							case "p":
							case "pre":
							case "section":
							case "summary":
							case "ul":
							case "button":
							case "label":
							case "menuitem":
								isHtmlTag = true;
								break;
							case "select":
								break;
							case "left":
							case "right":
							case "body":
								isHtmlTag = true;
								reactTag = "div";
								break;
							case "head":
							case "title":
							case "back":
								if(state.opts.platform !== "weixin"){
									isHtmlTag = true;
									reactTag = "div";
								}
								break;
							case "input":
							case "textarea":
							case "a":
								break;
							case "img":
								reactTag = "Image";
								break;
							case "area":
							case "audio":
							case "base":
							case "br":
							case "canvas":
							case "caption":
							case "col":
							case "colgroup":
							case "command":
							case "datalist":
							case "dialog":
							case "embed":
							case "frame":
							case "frameset":
							case "head":
							case "hr":
							case "html":
							case "iframe":
							case "keygen":
							case "link":
							case "map":
							case "meta":
							case "noframes":
							case "noscript":
							case "object":
							case "optgroup":
							case "option":
							case "param":
							case "progress":
							case "rp":
							case "rt":
							case "ruby":
							case "script":
							case "source":
							case "style":
							case "table":
							case "tbody":
							case "td":
							case "tfoot":
							case "th":
							case "thead":
							case "tr":
							case "track":
							case "video":
							case "wbr":
								isHtmlTag = true;
								console.error("标签 " + htmlTag + " 暂不支持");
								break;
							case "seo-t-d-k":
								if (state.opts.platform !== "seo") {
									path.remove();
								}
								break;
						}

						var classList = [];

						// 如果不是html标签，则转换为react标签
						if(!isHtmlTag){
							if(!reactTag){
								reactTag = htmlTag.replace(/(^[a-z]|\-[a-z])/g, function(char){
									return char.replace("-", "").toUpperCase();
								});
							}

							$arguments[0] = t.identifier(reactTag);
							
							classList.push("ej-tag-" + htmlTag.replace(/^[A-Z]/, function(char){
								return char.toLowerCase();
							}).replace(/[A-Z]/g, function(char){
								return "-" + char.toLowerCase();
							}));
						}else if(reactTag){
							$arguments[0] = t.stringLiteral(reactTag);
							
							classList.push("ej-tag-" + htmlTag);
						}

						// 检测生成属性节点
						let properties;

						if($arguments[1].type === "NullLiteral"){
							$arguments[1] = t.objectExpression([]);
							properties = $arguments[1].properties;
						}else if($arguments[1].type === "CallExpression" &&
							$arguments[1].callee.type === "Identifier" &&
							$arguments[1].callee.name === "_extends" &&
							$arguments[1].arguments.length){
							// 针对于<element {...xxx} />这种情况
							properties = [];
							$arguments[1].arguments.forEach(function(argument){
								if(argument.type === "ObjectExpression"){
									properties = properties.concat(argument.properties);
								}
							});
						}else if($arguments[1].type === "ObjectExpression"){
							properties = $arguments[1].properties;
						}else{
							console.error("属性类型咱不支持");
							console.log("bundle-loader-enjoy-rn-html/src/html-to-rn.js");
							console.log($arguments[1]);
						}

						properties.forEach(property => {
							var key = property.key;
							if(key.type === "Identifier"){
								key.name = toCamel(key.name);
							}else if(key.type === "StringLiteral"){
								key.value = toCamel(key.value);
							}
						});

						let componentProperty = properties.find(property => {
							return property.key.name === "component";
						});

						if(componentProperty){
							if(componentProperty.value.value){
								reactTag = toCamel(componentProperty.value.value).replace(/^[a-z]/, function(char){
									return char.toUpperCase();
								});
								
								$arguments[0] = t.identifier(reactTag);
								
								classList.push("ej-tag-" + componentProperty.value.value);
							}

							properties.splice(properties.indexOf(componentProperty), 1);
						}

						// 转换属性
						properties.forEach(property => {
							var key = property.key;
							switch(key.value){
								case "class":
									key.value = "className";
									break;
								case "for":
									key.value = "htmlFor";
									break;
							}
						});

						// 所有以style、Style结尾的都做转换
						let styleReg = /[sS]tyle$/;
						let stylePropertys = properties.filter(property => {
							return styleReg.test(property.key.value) || styleReg.test(property.key.name);
						});

						// 转换style
						stylePropertys.forEach((styleProperty) => {
							if(styleProperty.value.value){
								try{
									// 转换style为一个对象
									let styleObj = css2obj("*{" + styleProperty.value.value + "}", state.opts.platform)["*"];
									// 生成新的style节点
									let styleProperties = [];
									for(let key in styleObj){
										styleProperties.push(t.objectProperty(t.identifier(key), t.stringLiteral(styleObj[key])));
									}
									// 替换节点
									properties[properties.indexOf(styleProperty)] = t.objectProperty(t.identifier(styleProperty.key.name || styleProperty.key.value), t.objectExpression(styleProperties));
								}catch(e){
									console.error(e);
								}
							}
						});

						let classProperty = properties.find(property => {
							return property.key.value === "className";
						});

						// // 添加特殊class
						// switch(htmlTag){
						// 	case "head":
						// 		classList.push("navbar");
						// 		break;
						// 	case "title":
						// 		classList.push("navbar-title");
						// 		break;
						// 	case "left":
						// 		classList.push("navbar-left");
						// 		break;
						// 	case "right":
						// 		classList.push("navbar-right");
						// 		break;
						// }

						// if(classList.length){
						// 	classList = classList;
							if(!classProperty){
								classProperty = t.objectProperty(t.stringLiteral("className"), t.stringLiteral(classList.join(" ")));
								properties.push(classProperty);
							}else{
								if(classProperty.value.type === "StringLiteral"){
									classProperty.value.value += " " + classList.join(" ");
								}else if(classProperty.value.type !== "CallExpression" ||
									classProperty.value.callee.type !== "Identifier" ||
									classProperty.value.callee.name !== "__class__"){
									classList.push(classProperty.value);
									
									classProperty.value = t.callExpression(
																t.identifier("__class__"),
																[
																		t.arrayExpression(
																				classList.map(function(item){
																					return typeof item === "string" ? t.stringLiteral(item) : item;
																				})
																		)
																]);
								}
							}
						//}

						let isState = function(node){
							if(node.type === "MemberExpression"){
								if(node.object.type === "MemberExpression" &&
									node.object.object.type === "ThisExpression" &&
									node.object.property.type === "Identifier" &&
									node.object.property.name === "state"){
									return true;
								}else{
									return isState(node.object);
								}
							}else{
								return false;
							}
						};

						let getState = function(node){
							if(node.type === "MemberExpression"){
								if(node.object.type === "MemberExpression" &&
									node.object.object.type === "ThisExpression" &&
									node.object.property.type === "Identifier" &&
									node.object.property.name === "state"){
									return node.property;
								}else{
									return getState(node.object);
								}
							}else{
								return null;
							}
						};

						let twoWayBinding = function(node, properties){
							let onChangeProperty = properties.find(property => {
								return property.key.name === "onChange";
							});

							if(onChangeProperty){
								properties.splice(properties.indexOf(onChangeProperty), 1);
							}

							let stateKey = getState(node);

							let body = [
								t.expressionStatement(
									t.AssignmentExpression(
										"=",
										node,
										t.MemberExpression(
											t.MemberExpression(
												t.identifier("e"),
												t.identifier("target")
											),
											t.identifier("value")
										)
									)
								),
								t.expressionStatement(
									t.CallExpression(
										t.MemberExpression(
											t.thisExpression(),
											t.identifier("setState")
										),
										[
											t.ObjectExpression([
												t.objectProperty(
													stateKey,
													t.MemberExpression(
														t.MemberExpression(
															t.thisExpression(),
															t.identifier("state")
														),
														stateKey
													)
												)
											])
										]
									)
								)
							];

							if(onChangeProperty){
								body.push(t.expressionStatement(
									t.callExpression(
										onChangeProperty.value,
										[
											t.identifier("e")
										]
									)
								));
							}

							properties.push(t.ObjectProperty(
								t.identifier("onChange"),
								t.ArrowFunctionExpression(
									[t.identifier("e")],
									t.blockStatement(
										body,
										[]
									)
								)
							));
						};

						// 如果是导航组件，则植入_util工具
						if(reactTag === "Router"){
							properties.push(t.objectProperty(t.identifier("_util"), t.identifier("__util__")));
						}else if(reactTag === "A"){
							let hrefProperty = properties.find(property => {
								return property.key.value === "href" || property.key.name === "href";
							});
							if(hrefProperty && (hrefProperty.value.type !== "StringLiteral" || hrefProperty.value.value) && hrefProperty.value.type !== "NullLiteral"){
								properties.push(t.objectProperty(t.identifier("_util"), t.identifier("__util__")));
								properties.push(
									t.objectProperty(
										t.identifier("_navigator"),
										t.memberExpression(
											t.thisExpression(),
											t.identifier("page")
										)
									)
								);
							}
						}else if(reactTag === "Image"){
							let srcProperty = properties.find(property => {
								return property.key.value === "src" || property.key.name === "src";
							});

							if(srcProperty){
								if(srcProperty.value.type === "StringLiteral" && /^\.{1,2}\/[^\s]+\.(png|jpg|jpeg|gif)$/.test(srcProperty.value.value)){
									srcProperty.value = t.callExpression(t.identifier("require"), [srcProperty.value]);
								}else{
									properties.push(t.objectProperty(t.identifier("_util"), t.identifier("__util__")));
									properties.push(t.objectProperty(t.identifier("_inner_require"), t.identifier("__inner_require__")));
								}
							}
						}else if(reactTag === "Input"){
							let valueProperty = properties.find(property => {
								return property.key.name === "value";
							});

							if(valueProperty && isState(valueProperty.value)){
								twoWayBinding(valueProperty.value, properties);
							}
						}else if(reactTag === "Textarea"){
							let valueProperty = properties.find(property => {
								return property.key.name === "value";
							});
							if(valueProperty){
								properties.splice(properties.indexOf(valueProperty), 1);
							}
							if($arguments[2]){
								if(isState($arguments[2])){
									twoWayBinding($arguments[2], properties);
								}else{
									properties.push(
										t.objectProperty(
											t.identifier("value"),
											$arguments[2]
										)
									);
								}
							}
						}

						// 自定义标签，要将导航传入
						if(path.scope.hasBinding(reactTag)){
							properties.push(t.objectProperty(t.identifier("_pageInfo"),
															t.memberExpression(
																			t.memberExpression(t.thisExpression(),
																			t.identifier("props")
															), t.identifier("_pageInfo"))));
						}

						let refProperty = properties.find(property => {
							return property.key.value === "ref" || property.key.name === "ref";
						});
						// 解析导航栏节点中的ref属性
						if(refProperty){
							let tempPath = path;
							while(tempPath = tempPath.parentPath){
								let tempNode = tempPath.node;
								if(tempNode.type === "CallExpression" &&
									(tempNode = tempNode.callee) &&
									tempNode.type === "MemberExpression" &&
									tempNode.property.type === "Identifier" &&
									tempNode.property.name === "view" &&
									(tempNode = tempNode.object) &&
									tempNode.type === "MemberExpression" &&
									tempNode.property.type === "Identifier" &&
									tempNode.property.name === "navbar" &&
									(tempNode = tempNode.object) &&
									tempNode.type === "MemberExpression" &&
									tempNode.property.type === "Identifier" &&
									tempNode.property.name === "page" &&
									tempNode.object.type === "ThisExpression"){
									// ref的key
									let key = refProperty.value.value;
									// 删除ref属性
									properties.splice(properties.indexOf(refProperty), 1);
									// 添加_ref属性
									// _ref = self => { this.refs.xxx = self; }
									properties.push(t.objectProperty(
													t.identifier("_ref"),
													t.arrowFunctionExpression(
															[
																	t.identifier("self")
															],
															t.blockStatement(
																	[
																			t.expressionStatement(
																					t.assignmentExpression(
																							"=",
																							t.memberExpression(
																									t.memberExpression(
																											t.thisExpression(),
																											t.identifier("_refs")
																									),
																									t.identifier(key)
																							),
																							t.identifier("self")
																					)
																			)
																	],
																	[]
															)
													)
											));
									break;
								}
							}
						}

						// let onLayoutProperty = properties.find(property => {
						// 	return property.key.name === "onLayout";
						// });

						// if(onLayoutProperty){
						// 	properties.push(t.objectProperty(
						// 			t.identifier("onCompositionEnd"),
						// 			onLayoutProperty.value
						// 	));

						// 	properties.splice(properties.indexOf(onLayoutProperty), 1);
						// }

						switch(htmlTag){
							// 对于返回按钮，自动添加click事件，处理导航的后退功能
							case "back":
								let body = [
									t.expressionStatement(
										t.callExpression(
											t.memberExpression(
													t.identifier("__nav_util__"),
													t.identifier("back")
											),
											[]
										)
									)
								];

								let clickProperty = properties.find(property => {
									return property.key.name === "onClick";
								});

								if(clickProperty){
									body.push(
										t.expressionStatement(
											t.callExpression(
												clickProperty.value,
												[]
											)
										)
									);
									properties.splice(properties.indexOf(clickProperty), 1);
								}

								properties.push(t.objectProperty(
														t.identifier("onClick"),
														t.arrowFunctionExpression(
															[],
															t.blockStatement(
																body,
																[]
															)
														)
												));
								path.replaceWith(t.conditionalExpression(
										t.memberExpression(
												t.identifier("__nav_util__"),
												t.identifier("canBack")
										),
										node,
										t.NullLiteral()
								));
								break;
						}
					}
				}
			},
			// 删除生成文件头部的"use strict"
			Directive: {
				enter(path){
					if(path.node.value.value === "use strict"){
						path.remove();
					}
				}
			}
		}
	};
};